@@ -4,7 +4,9 @@ import glob
 import json
 import logging
 import os
+import pickle
 import shutil
+import subprocess
 from copy import deepcopy
 from pathlib import Path
 from unittest import mock
@@ -13,18 +15,13 @@ import pytest
 
 from sqlfluff.cli.commands import lint
 from sqlfluff.core import FluffConfig, Lexer, Linter
-from sqlfluff.core.errors import SQLFluffSkipFile
+from sqlfluff.core.errors import SQLFluffSkipFile, SQLFluffUserError, SQLTemplaterError
 from sqlfluff.utils.testing.cli import invoke_assert_code
 from sqlfluff.utils.testing.logging import fluff_log_catcher
 from sqlfluff_templater_dbt.templater import DbtTemplater
-from test.fixtures.dbt.templater import (  # noqa: F401
-    DBT_FLUFF_CONFIG,
-    dbt_templater,
-    project_dir,
-)
 
 
-def test__templater_dbt_missing(dbt_templater, project_dir):  # noqa: F811
+def test__templater_dbt_missing(dbt_templater, project_dir, dbt_fluff_config):
     """Check that a nice error is returned when dbt module is missing."""
     try:
         import dbt  # noqa: F401
@@ -37,11 +34,11 @@ def test__templater_dbt_missing(dbt_templater, project_dir):  # noqa: F811
         dbt_templater.process(
             in_str="",
             fname=os.path.join(project_dir, "models/my_new_project/test.sql"),
-            config=FluffConfig(configs=DBT_FLUFF_CONFIG),
+            config=FluffConfig(configs=dbt_fluff_config),
         )
 
 
-def test__templater_dbt_profiles_dir_expanded(dbt_templater):  # noqa: F811
+def test__templater_dbt_profiles_dir_expanded(dbt_templater):
     """Check that the profiles_dir is expanded."""
     dbt_templater.sqlfluff_config = FluffConfig(
         configs={
@@ -91,35 +88,59 @@ def test__templater_dbt_profiles_dir_expanded(dbt_templater):  # noqa: F811
     ],
 )
 def test__templater_dbt_templating_result(
-    project_dir, dbt_templater, fname  # noqa: F811
+    project_dir,
+    dbt_templater,
+    fname,
+    dbt_fluff_config,
+    dbt_project_folder,
 ):
     """Test that input sql file gets templated into output sql file."""
-    _run_templater_and_verify_result(dbt_templater, project_dir, fname)
+    _run_templater_and_verify_result(
+        dbt_templater,
+        project_dir,
+        fname,
+        dbt_fluff_config,
+        dbt_project_folder,
+    )
 
 
 def test_dbt_profiles_dir_env_var_uppercase(
-    project_dir, dbt_templater, tmpdir, monkeypatch  # noqa: F811
+    project_dir,
+    dbt_templater,
+    tmpdir,
+    monkeypatch,
+    dbt_fluff_config,
+    dbt_project_folder,
+    profiles_dir,
 ):
     """Tests specifying the dbt profile dir with env var."""
-    profiles_dir = tmpdir.mkdir("SUBDIR")  # Use uppercase to test issue 2253
-    monkeypatch.setenv("DBT_PROFILES_DIR", str(profiles_dir))
-    shutil.copy(
-        os.path.join(project_dir, "../profiles_yml/profiles.yml"), str(profiles_dir)
+    sub_profiles_dir = tmpdir.mkdir("SUBDIR")  # Use uppercase to test issue 2253
+    monkeypatch.setenv("DBT_PROFILES_DIR", str(sub_profiles_dir))
+    shutil.copy(os.path.join(profiles_dir, "profiles.yml"), str(sub_profiles_dir))
+    _run_templater_and_verify_result(
+        dbt_templater,
+        project_dir,
+        "use_dbt_utils.sql",
+        dbt_fluff_config,
+        dbt_project_folder,
     )
-    _run_templater_and_verify_result(dbt_templater, project_dir, "use_dbt_utils.sql")
 
 
-def _run_templater_and_verify_result(dbt_templater, project_dir, fname):  # noqa: F811
+def _run_templater_and_verify_result(
+    dbt_templater,
+    project_dir,
+    fname,
+    dbt_fluff_config,
+    dbt_project_folder,
+):
     path = Path(project_dir) / "models/my_new_project" / fname
-    config = FluffConfig(configs=DBT_FLUFF_CONFIG)
+    config = FluffConfig(configs=dbt_fluff_config)
     templated_file, _ = dbt_templater.process(
         in_str=path.read_text(),
         fname=str(path),
         config=config,
     )
-    template_output_folder_path = Path(
-        "plugins/sqlfluff-templater-dbt/test/fixtures/dbt/templated_output/"
-    )
+    template_output_folder_path = dbt_project_folder / "templated_output/"
     fixture_path = _get_fixture_path(template_output_folder_path, fname)
     assert str(templated_file) == fixture_path.read_text()
     # Check we can lex the output too.
@@ -180,12 +201,16 @@ def _get_fixture_path(template_output_folder_path, fname):
     ],
 )
 def test__templater_dbt_sequence_files_ephemeral_dependency(
-    project_dir, dbt_templater, fnames_input, fnames_expected_sequence  # noqa: F811
+    project_dir,
+    dbt_templater,
+    fnames_input,
+    fnames_expected_sequence,
+    dbt_fluff_config,
 ):
     """Test that dbt templater sequences files based on dependencies."""
     result = dbt_templater.sequence_files(
         [str(Path(project_dir) / fn) for fn in fnames_input],
-        config=FluffConfig(configs=DBT_FLUFF_CONFIG),
+        config=FluffConfig(configs=dbt_fluff_config),
     )
     pd = Path(project_dir)
     expected = [str(pd / fn) for fn in fnames_expected_sequence]
@@ -210,7 +235,11 @@ select * from a
     ],
 )
 def test__templater_dbt_slice_file_wrapped_test(
-    raw_file, templated_file, result, dbt_templater, caplog  # noqa: F811
+    raw_file,
+    templated_file,
+    result,
+    dbt_templater,
+    caplog,
 ):
     """Test that wrapped queries are sliced safely using _check_for_wrapped()."""
 
@@ -242,14 +271,17 @@ def test__templater_dbt_slice_file_wrapped_test(
     ],
 )
 def test__templater_dbt_templating_test_lex(
-    project_dir, dbt_templater, fname  # noqa: F811
+    project_dir,
+    dbt_templater,
+    fname,
+    dbt_fluff_config,
 ):
     """Demonstrate the lexer works on both dbt models and dbt tests.
 
     Handle any number of newlines.
     """
     path = Path(project_dir) / fname
-    config = FluffConfig(configs=DBT_FLUFF_CONFIG)
+    config = FluffConfig(configs=dbt_fluff_config)
     source_dbt_sql = path.read_text()
     # Count the newlines.
     n_trailing_newlines = len(source_dbt_sql) - len(source_dbt_sql.rstrip("\n"))
@@ -291,14 +323,28 @@ def test__templater_dbt_templating_test_lex(
     ],
 )
 def test__templater_dbt_skips_file(
-    path, reason, dbt_templater, project_dir  # noqa: F811
+    path,
+    reason,
+    dbt_templater,
+    project_dir,
+    dbt_fluff_config,
 ):
     """A disabled dbt model should be skipped."""
     with pytest.raises(SQLFluffSkipFile, match=reason):
         dbt_templater.process(
             in_str="",
             fname=os.path.join(project_dir, path),
-            config=FluffConfig(configs=DBT_FLUFF_CONFIG),
+            config=FluffConfig(configs=dbt_fluff_config),
+        )
+
+
+def test_dbt_fails_stdin(dbt_templater, dbt_fluff_config):
+    """Reading from stdin is not supported with dbt templater."""
+    with pytest.raises(SQLFluffUserError):
+        dbt_templater.process(
+            in_str="",
+            fname="stdin",
+            config=FluffConfig(configs=dbt_fluff_config),
         )
 
 
@@ -312,10 +358,13 @@ def test__templater_dbt_skips_file(
     ],
 )
 def test__dbt_templated_models_do_not_raise_lint_error(
-    project_dir, fname, caplog  # noqa: F811
+    project_dir,
+    fname,
+    caplog,
+    dbt_fluff_config,
 ):
     """Test that templated dbt models do not raise a linting error."""
-    linter = Linter(config=FluffConfig(configs=DBT_FLUFF_CONFIG))
+    linter = Linter(config=FluffConfig(configs=dbt_fluff_config))
     # Log rules output.
     with caplog.at_level(logging.DEBUG, logger="sqlfluff.rules"):
         lnt = linter.lint_path(
@@ -330,7 +379,7 @@ def test__dbt_templated_models_do_not_raise_lint_error(
         print(linted_file.tree.stringify())
         print("\n\n## VIOLATIONS:")
         for idx, v in enumerate(linted_file.violations):
-            print(f"   {idx}:{v.get_info_dict()}")
+            print(f"   {idx}:{v.to_dict()}")
 
     violations = lnt.check_tuples()
     assert len(violations) == 0
@@ -346,12 +395,15 @@ def _clean_path(glob_expression):
     "path", ["models/my_new_project/issue_1608.sql", "snapshots/issue_1771.sql"]
 )
 def test__dbt_templated_models_fix_does_not_corrupt_file(
-    project_dir, path, caplog  # noqa: F811
+    project_dir,
+    path,
+    caplog,
+    dbt_fluff_config,
 ):
     """Test issues where previously "sqlfluff fix" corrupted the file."""
     test_glob = os.path.join(project_dir, os.path.dirname(path), "*FIXED.sql")
     _clean_path(test_glob)
-    lntr = Linter(config=FluffConfig(configs=DBT_FLUFF_CONFIG))
+    lntr = Linter(config=FluffConfig(configs=dbt_fluff_config))
     with caplog.at_level(logging.INFO, logger="sqlfluff.linter"):
         lnt = lntr.lint_path(os.path.join(project_dir, path), fix=True)
     try:
@@ -366,7 +418,9 @@ def test__dbt_templated_models_fix_does_not_corrupt_file(
 
 
 def test__templater_dbt_templating_absolute_path(
-    project_dir, dbt_templater  # noqa: F811
+    project_dir,
+    dbt_templater,
+    dbt_fluff_config,
 ):
     """Test that absolute path of input path does not cause RuntimeError."""
     try:
@@ -375,21 +429,22 @@ def test__templater_dbt_templating_absolute_path(
             fname=os.path.abspath(
                 os.path.join(project_dir, "models/my_new_project/use_var.sql")
             ),
-            config=FluffConfig(configs=DBT_FLUFF_CONFIG),
+            config=FluffConfig(configs=dbt_fluff_config),
         )
     except Exception as e:
         pytest.fail(f"Unexpected RuntimeError: {e}")
 
 
 @pytest.mark.parametrize(
-    "fname,exception_msg",
+    "fname,exception_msg,exception_class",
     [
         (
             "compiler_error.sql",
-            "dbt compilation error on file 'models/my_new_project/compiler_error.sql'"
-            ", Unexpected end of template. Jinja was looking for the following tags:"
-            " 'endfor' or 'else'. The innermost block that needs to be closed is "
-            "'for'.\n  line 5\n    {{ col }}",
+            "Compilation Error in model compiler_error "
+            "(models/my_new_project/compiler_error.sql)\n  "
+            "Unexpected end of template. Jinja was looking for the following tags: "
+            "'endfor' or 'else'.",
+            SQLFluffUserError,
         ),
         (
             "unknown_ref.sql",
@@ -397,6 +452,7 @@ def test__templater_dbt_templating_absolute_path(
             "Model 'model.my_new_project.unknown_ref' "
             "(models/my_new_project/unknown_ref.sql) depends on a node named "
             "'i_do_not_exist' which was not found",
+            SQLFluffUserError,
         ),
         (
             "unknown_macro.sql",
@@ -404,51 +460,99 @@ def test__templater_dbt_templating_absolute_path(
             "Compilation Error in model unknown_macro "
             "(models/my_new_project/unknown_macro.sql)\n  'invalid_macro' is "
             "undefined. This can happen when calling a macro that does not exist.",
+            SQLTemplaterError,
+        ),
+        (
+            "compile_missing_table.sql",
+            # In the test suite we don't get a very helpful error message from dbt
+            # but in live testing, the inclusion of the triggering error sometimes
+            # gives us something much more useful.
+            "because dbt raised a fatal exception during compilation",
+            SQLFluffSkipFile,
         ),
     ],
 )
 def test__templater_dbt_handle_exceptions(
-    project_dir, dbt_templater, fname, exception_msg  # noqa: F811
+    project_dir,
+    dbt_templater,
+    dbt_fluff_config,
+    dbt_project_folder,
+    fname,
+    exception_msg,
+    exception_class,
 ):
     """Test that exceptions during compilation are returned as violation."""
     from dbt.adapters.factory import get_adapter
 
-    src_fpath = "plugins/sqlfluff-templater-dbt/test/fixtures/dbt/error_models/" + fname
+    src_fpath = dbt_project_folder / "error_models" / fname
     target_fpath = os.path.abspath(
         os.path.join(project_dir, "models/my_new_project/", fname)
     )
     # We move the file that throws an error in and out of the project directory
     # as dbt throws an error if a node fails to parse while computing the DAG
-    os.rename(src_fpath, target_fpath)
+    shutil.move(src_fpath, target_fpath)
     try:
-        _, violations = dbt_templater.process(
-            in_str="",
-            fname=target_fpath,
-            config=FluffConfig(configs=DBT_FLUFF_CONFIG, overrides={"dialect": "ansi"}),
-        )
+        with pytest.raises(exception_class) as excinfo:
+            dbt_templater.process(
+                in_str="",
+                fname=target_fpath,
+                config=FluffConfig(
+                    configs=dbt_fluff_config, overrides={"dialect": "ansi"}
+                ),
+            )
     finally:
-        os.rename(target_fpath, src_fpath)
+        shutil.move(target_fpath, src_fpath)
         get_adapter(dbt_templater.dbt_config).connections.release()
-    assert violations
+
+    # Debug logging.
+    print("Raised:", excinfo.value)
+    for trace in excinfo.traceback:
+        print(trace)
+
     # NB: Replace slashes to deal with different platform paths being returned.
-    assert exception_msg in violations[0].desc().replace("\\", "/")
+    if exception_class is SQLTemplaterError:
+        _msg = excinfo.value.desc().replace("\\", "/")
+    else:
+        _msg = str(excinfo.value).replace("\\", "/")
+    assert exception_msg in _msg
+    # Ensure that there's no context parent exception, because they don't pickle well.
+    # https://github.com/sqlfluff/sqlfluff/issues/6037
+    # We *should* be stripping any inherited exceptions from anything returned here.
+    # Any residual dbt exceptions are a risk for pickling errors.
+    assert not excinfo.value.__context__
+    assert not excinfo.value.__cause__
+    # We also ensure that the exception can be pickled and unpickled safely.
+    # Pickling of exceptions happens during parallel operation and so if it can't
+    # be done safely then that will cause bugs.
+    pickled_exception = pickle.dumps(excinfo.value)
+    roundtrip_exception = pickle.loads(pickled_exception)
+    assert isinstance(roundtrip_exception, type(excinfo.value))
+    assert str(roundtrip_exception) == str(excinfo.value)
 
 
 @mock.patch("dbt.adapters.postgres.impl.PostgresAdapter.set_relations_cache")
 def test__templater_dbt_handle_database_connection_failure(
-    set_relations_cache, project_dir, dbt_templater  # noqa: F811
+    set_relations_cache,
+    project_dir,
+    dbt_templater,
+    dbt_fluff_config,
 ):
     """Test the result of a failed database connection."""
     from dbt.adapters.factory import get_adapter
 
     try:
-        from dbt.exceptions import (
-            FailedToConnectException as DbtFailedToConnectException,
-        )
-    except ImportError:
-        from dbt.exceptions import (
+        from dbt.adapters.exceptions import (
             FailedToConnectError as DbtFailedToConnectException,
         )
+    except ImportError:
+        try:
+            from dbt.exceptions import (
+                FailedToConnectError as DbtFailedToConnectException,
+            )
+        except ImportError:
+            from dbt.exceptions import (
+                FailedToConnectException as DbtFailedToConnectException,
+            )
 
     # Clear the adapter cache to force this test to create a new connection.
     DbtTemplater.adapters.clear()
@@ -464,33 +568,29 @@ def test__templater_dbt_handle_database_connection_failure(
             project_dir, "models/my_new_project/exception_connect_database.sql"
         )
     )
-    dbt_fluff_config_fail = deepcopy(DBT_FLUFF_CONFIG)
+    dbt_fluff_config_fail = deepcopy(dbt_fluff_config)
     dbt_fluff_config_fail["templater"]["dbt"][
         "profiles_dir"
     ] = "plugins/sqlfluff-templater-dbt/test/fixtures/dbt/profiles_yml_fail"
     # We move the file that throws an error in and out of the project directory
     # as dbt throws an error if a node fails to parse while computing the DAG
-    os.rename(src_fpath, target_fpath)
+    shutil.move(src_fpath, target_fpath)
     try:
-        _, violations = dbt_templater.process(
-            in_str="",
-            fname=target_fpath,
-            config=FluffConfig(configs=DBT_FLUFF_CONFIG),
-        )
+        with pytest.raises(SQLTemplaterError) as excinfo:
+            dbt_templater.process(
+                in_str="",
+                fname=target_fpath,
+                config=FluffConfig(configs=dbt_fluff_config),
+            )
     finally:
-        os.rename(target_fpath, src_fpath)
+        shutil.move(target_fpath, src_fpath)
         get_adapter(dbt_templater.dbt_config).connections.release()
-    assert violations
     # NB: Replace slashes to deal with different platform paths being returned.
-    assert (
-        violations[0]
-        .desc()
-        .replace("\\", "/")
-        .startswith("dbt tried to connect to the database")
-    )
+    error_message = excinfo.value.desc().replace("\\", "/")
+    assert "dbt tried to connect to the database" in error_message
 
 
-def test__project_dir_does_not_exist_error(dbt_templater):  # noqa: F811
+def test__project_dir_does_not_exist_error(dbt_templater):
     """Test an error is logged if the given dbt project directory doesn't exist."""
     dbt_templater.sqlfluff_config = FluffConfig(
         configs={
@@ -514,12 +614,16 @@ def test__project_dir_does_not_exist_error(dbt_templater):  # noqa: F811
     ],
 )
 def test__context_in_config_is_loaded(
-    project_dir, dbt_templater, model_path, var_value  # noqa: F811
+    project_dir,
+    dbt_templater,
+    model_path,
+    var_value,
+    dbt_fluff_config,
 ):
     """Test that variables inside .sqlfluff are passed to dbt."""
     context = {"passed_through_cli": var_value} if var_value else {}
 
-    config_dict = deepcopy(DBT_FLUFF_CONFIG)
+    config_dict = deepcopy(dbt_fluff_config)
     config_dict["templater"]["dbt"]["context"] = context
     config = FluffConfig(config_dict)
 
@@ -533,7 +637,34 @@ def test__context_in_config_is_loaded(
     assert str(var_value) in processed.templated_str
 
 
-def test__dbt_log_supression():
+@pytest.mark.parametrize(
+    ("model_path", "var_value"),
+    [
+        ("models/vars_from_env.sql", "expected_value"),
+    ],
+)
+def test__context_in_env_is_loaded(
+    project_dir,
+    dbt_templater,
+    model_path,
+    var_value,
+    dbt_fluff_config,
+):
+    """Test that variables inside env are passed to dbt."""
+    os.environ["passed_through_env"] = var_value
+
+    config = FluffConfig(dbt_fluff_config)
+    path = Path(project_dir) / model_path
+
+    processed, violations = dbt_templater.process(
+        in_str=path.read_text(), fname=str(path), config=config
+    )
+
+    assert violations == []
+    assert str(var_value) in processed.templated_str
+
+
+def test__dbt_log_supression(dbt_project_folder):
     """Test that when we try and parse in JSON format we get JSON.
 
     This actually tests that we can successfully suppress unwanted
@@ -541,23 +672,32 @@ def test__dbt_log_supression():
     """
     oldcwd = os.getcwd()
     try:
-        os.chdir("plugins/sqlfluff-templater-dbt/test/fixtures/dbt")
+        os.chdir(dbt_project_folder)
+
+        cli_options = [
+            "--disable-progress-bar",
+            "dbt_project/models/my_new_project/operator_errors.sql",
+            "-f",
+            "json",
+        ]
+
         result = invoke_assert_code(
             ret_code=1,
             args=[
                 lint,
-                [
-                    "--disable-progress-bar",
-                    "dbt_project/models/my_new_project/operator_errors.sql",
-                    "-f",
-                    "json",
-                ],
+                cli_options,
             ],
         )
+        # the CliRunner isn't isolated from the dbt plugin loading
+        isolated_lint = subprocess.run(
+            ["sqlfluff", "lint"] + cli_options, capture_output=True
+        )
     finally:
         os.chdir(oldcwd)
     # Check that the full output parses as json
     parsed = json.loads(result.output)
+    assert isolated_lint.returncode == 1
+    assert b" Registered adapter:" not in isolated_lint.stdout
     assert isinstance(parsed, list)
     assert len(parsed) == 1
     first_file = parsed[0]
