@@ -8,6 +8,7 @@ default keyword definition, these keywords are, or have been, an ANSI keyword.
 There are also some keywords that are(n't) supported as types and function, but there
 isn't support for that distinction at present.
 """
+
 from typing import List, Tuple
 
 
@@ -410,6 +411,7 @@ postgres_docs_keywords = [
     ("INVOKER", "non-reserved"),
     ("IS", "reserved-(can-be-function-or-type)"),
     ("ISNULL", "reserved-(can-be-function-or-type)"),
+    ("IS_SUPERUSER", "non-reserved"),
     ("ISOLATION", "non-reserved"),
     ("JOIN", "reserved-(can-be-function-or-type)"),
     ("JSON", "not-keyword"),
@@ -436,6 +438,8 @@ postgres_docs_keywords = [
     ("LAST", "non-reserved"),
     ("LAST_VALUE", "not-keyword"),
     ("LATERAL", "reserved"),
+    ("LC_COLLATE", "non-reserved"),
+    ("LC_CTYPE", "non-reserved"),
     ("LEAD", "not-keyword"),
     ("LEADING", "reserved"),
     ("LEAKPROOF", "non-reserved"),
@@ -663,7 +667,8 @@ postgres_docs_keywords = [
     ("RESTRICT", "non-reserved"),
     ("RESTRICTIVE", "non-reserved"),
     ("RESULT", "not-keyword"),
-    ("RETURN", "not-keyword"),
+    ("RETRIEVE", "non-reserved"),
+    ("RETURN", "non-reserved"),
     ("RETURNED_CARDINALITY", "not-keyword"),
     ("RETURNED_LENGTH", "not-keyword"),
     ("RETURNED_OCTET_LENGTH", "not-keyword"),
@@ -714,7 +719,9 @@ postgres_docs_keywords = [
     ("SERIAL8", "non-reserved-(cannot-be-function-or-type)"),
     ("SERIALIZABLE", "non-reserved"),
     ("SERVER", "non-reserved"),
+    ("SERVER_ENCODING", "non-reserved"),
     ("SERVER_NAME", "not-keyword"),
+    ("SERVER_VERSION", "non-reserved"),
     ("SESSION", "non-reserved"),
     ("SESSION_USER", "reserved"),
     ("SET", "non-reserved"),
@@ -925,6 +932,7 @@ postgres_nondocs_keywords = [
     ("CREATEDB", "non-reserved"),
     ("CREATEROLE", "non-reserved"),
     ("DATE", "non-reserved"),
+    ("DEPENDENCIES", "non-reserved"),
     ("DEPTH", "non-reserved"),
     ("DESCRIBE", "non-reserved"),
     ("DETERMINISTIC", "non-reserved"),
@@ -949,7 +957,9 @@ postgres_nondocs_keywords = [
     ("LOGIN", "non-reserved"),
     ("LOCALE", "non-reserved"),
     ("MAIN", "non-reserved"),
+    ("MCV", "non-reserved"),
     ("MODULUS", "non-reserved"),
+    ("NDISTINCT", "non-reserved"),
     ("NOBYPASSRLS", "non-reserved"),
     ("NOCREATEDB", "non-reserved"),
     ("NOCREATEROLE", "non-reserved"),
@@ -1023,9 +1033,14 @@ postgres_postgis_other_keywords = [
     ("EMPTY", "non-reserved"),
 ]
 
+postgres_pgvector_keywords = [
+    ("VECTOR", "non-reserved"),
+]
+
 postgres_keywords = priority_keyword_merge(
     postgres_docs_keywords,
     postgres_nondocs_keywords,
     postgres_postgis_datatype_keywords,
     postgres_postgis_other_keywords,
+    postgres_pgvector_keywords,
 )
