@@ -12,11 +12,13 @@ from sqlfluff.core import FluffConfig, Linter
 from sqlfluff.core.errors import (
     SQLBaseError,
     SQLFluffSkipFile,
-    SQLFluffUserError,
     SQLLexError,
     SQLLintError,
+    SQLParseError,
+    SQLTemplaterError,
 )
-from sqlfluff.core.linter import LintingResult, runner
+from sqlfluff.core.linter import runner
+from sqlfluff.core.linter.linting_result import combine_dicts, sum_dicts
 from sqlfluff.core.linter.runner import get_runner
 from sqlfluff.utils.testing.logging import fluff_log_catcher
 
@@ -38,44 +40,6 @@ def normalise_paths(paths):
     return {pth.replace("/", ".").replace("\\", ".") for pth in paths}
 
 
-def test__linter__path_from_paths__dir():
-    """Test extracting paths from directories."""
-    lntr = Linter()
-    paths = lntr.paths_from_path("test/fixtures/lexer")
-    assert normalise_paths(paths) == {
-        "test.fixtures.lexer.block_comment.sql",
-        "test.fixtures.lexer.inline_comment.sql",
-        "test.fixtures.lexer.basic.sql",
-    }
-
-
-def test__linter__path_from_paths__default():
-    """Test .sql files are found by default."""
-    lntr = Linter()
-    paths = normalise_paths(lntr.paths_from_path("test/fixtures/linter"))
-    assert "test.fixtures.linter.passing.sql" in paths
-    assert "test.fixtures.linter.passing_cap_extension.SQL" in paths
-    assert "test.fixtures.linter.discovery_file.txt" not in paths
-
-
-def test__linter__path_from_paths__exts():
-    """Test configuration of file discovery."""
-    lntr = Linter(
-        config=FluffConfig(overrides={"sql_file_exts": ".txt", "dialect": "ansi"})
-    )
-    paths = normalise_paths(lntr.paths_from_path("test/fixtures/linter"))
-    assert "test.fixtures.linter.passing.sql" not in paths
-    assert "test.fixtures.linter.passing_cap_extension.SQL" not in paths
-    assert "test.fixtures.linter.discovery_file.txt" in paths
-
-
-def test__linter__path_from_paths__file():
-    """Test extracting paths from a file path."""
-    lntr = Linter()
-    paths = lntr.paths_from_path("test/fixtures/linter/indentation_errors.sql")
-    assert normalise_paths(paths) == {"test.fixtures.linter.indentation_errors.sql"}
-
-
 @pytest.mark.parametrize("filesize,raises_skip", [(0, False), (5, True), (2000, False)])
 def test__linter__skip_large_bytes(filesize, raises_skip):
     """Test extracting paths from a file path."""
@@ -114,79 +78,6 @@ def test__linter__skip_large_bytes(filesize, raises_skip):
         assert result
 
 
-def test__linter__path_from_paths__not_exist():
-    """Test that the right errors are raise when a file doesn't exist."""
-    lntr = Linter()
-    with pytest.raises(SQLFluffUserError):
-        lntr.paths_from_path("asflekjfhsakuefhse")
-
-
-def test__linter__path_from_paths__not_exist_ignore():
-    """Test extracting paths from a file path."""
-    lntr = Linter()
-    paths = lntr.paths_from_path("asflekjfhsakuefhse", ignore_non_existent_files=True)
-    assert len(paths) == 0
-
-
-def test__linter__path_from_paths__explicit_ignore():
-    """Test ignoring files that were passed explicitly."""
-    lntr = Linter()
-    paths = lntr.paths_from_path(
-        "test/fixtures/linter/sqlfluffignore/path_a/query_a.sql",
-        ignore_non_existent_files=True,
-        ignore_files=True,
-        working_path="test/fixtures/linter/sqlfluffignore/",
-    )
-    assert len(paths) == 0
-
-
-def test__linter__path_from_paths__sqlfluffignore_current_directory():
-    """Test that .sqlfluffignore in the current directory is read when dir given."""
-    oldcwd = os.getcwd()
-    try:
-        os.chdir("test/fixtures/linter/sqlfluffignore")
-        lntr = Linter()
-        paths = lntr.paths_from_path(
-            "path_a/",
-            ignore_non_existent_files=True,
-            ignore_files=True,
-            working_path="test/fixtures/linter/sqlfluffignore/",
-        )
-        assert len(paths) == 0
-    finally:
-        os.chdir(oldcwd)
-
-
-def test__linter__path_from_paths__dot():
-    """Test extracting paths from a dot."""
-    lntr = Linter()
-    paths = lntr.paths_from_path(".")
-    # Use set theory to check that we get AT LEAST these files
-    assert normalise_paths(paths) >= {
-        "test.fixtures.lexer.block_comment.sql",
-        "test.fixtures.lexer.inline_comment.sql",
-        "test.fixtures.lexer.basic.sql",
-    }
-
-
-@pytest.mark.parametrize(
-    "path",
-    [
-        "test/fixtures/linter/sqlfluffignore",
-        "test/fixtures/linter/sqlfluffignore/",
-        "test/fixtures/linter/sqlfluffignore/.",
-    ],
-)
-def test__linter__path_from_paths__ignore(path):
-    """Test extracting paths from a dot."""
-    lntr = Linter()
-    paths = lntr.paths_from_path(path)
-    # We should only get query_b, because of the sqlfluffignore files.
-    assert normalise_paths(paths) == {
-        "test.fixtures.linter.sqlfluffignore.path_b.query_b.sql"
-    }
-
-
 @pytest.mark.parametrize(
     "path",
     [
@@ -217,39 +108,131 @@ def test__linter__get_violations_filter_rules(rules, num_violations):
 
 def test__linter__linting_result__sum_dicts():
     """Test the summing of dictionaries in the linter."""
-    lr = LintingResult()
     i = {}
     a = dict(a=3, b=123, f=876.321)
     b = dict(a=19, b=321.0, g=23478)
     r = dict(a=22, b=444.0, f=876.321, g=23478)
-    assert lr.sum_dicts(a, b) == r
+    assert sum_dicts(a, b) == r
     # Check the identity too
-    assert lr.sum_dicts(r, i) == r
+    assert sum_dicts(r, i) == r
 
 
 def test__linter__linting_result__combine_dicts():
     """Test the combination of dictionaries in the linter."""
-    lr = LintingResult()
     a = dict(a=3, b=123, f=876.321)
     b = dict(h=19, i=321.0, j=23478)
     r = dict(z=22)
-    assert lr.combine_dicts(a, b, r) == dict(
+    assert combine_dicts(a, b, r) == dict(
         a=3, b=123, f=876.321, h=19, i=321.0, j=23478, z=22
     )
 
 
-@pytest.mark.parametrize("by_path,result_type", [(False, list), (True, dict)])
-def test__linter__linting_result_check_tuples_by_path(by_path, result_type):
+def test__linter__linting_result_check_tuples():
     """Test that a LintingResult can partition violations by the source files."""
     lntr = Linter()
     result = lntr.lint_paths(
-        [
+        (
             "test/fixtures/linter/comma_errors.sql",
             "test/fixtures/linter/whitespace_errors.sql",
-        ]
+        )
     )
-    check_tuples = result.check_tuples(by_path=by_path)
-    isinstance(check_tuples, result_type)
+    check_tuples = result.check_tuples()
+    isinstance(check_tuples, list)
+    assert check_tuples == [
+        ("LT09", 2, 1),
+        ("LT04", 4, 5),
+        ("LT02", 5, 1),
+        ("LT04", 5, 1),
+        ("LT02", 6, 1),
+        ("AL02", 6, 5),
+        ("LT01", 6, 6),
+        ("CP01", 8, 1),
+        ("LT09", 1, 1),
+        ("LT01", 2, 9),
+        ("LT01", 3, 12),
+        ("LT02", 4, 1),
+        ("CP01", 6, 10),
+    ]
+
+
+def test__linter__linting_result_check_tuples_by_path():
+    """Test that a LintingResult can partition violations by the source files."""
+    lntr = Linter()
+    result = lntr.lint_paths(
+        (
+            "test/fixtures/linter/comma_errors.sql",
+            "test/fixtures/linter/whitespace_errors.sql",
+        )
+    )
+    check_tuples = result.check_tuples_by_path()
+    isinstance(check_tuples, dict)
+    # Normalise the paths in the keys.
+    check_tuples = {k.replace("\\", "/"): v for k, v in check_tuples.items()}
+    assert check_tuples == {
+        "test/fixtures/linter/comma_errors.sql": [
+            ("LT09", 2, 1),
+            ("LT04", 4, 5),
+            ("LT02", 5, 1),
+            ("LT04", 5, 1),
+            ("LT02", 6, 1),
+            ("AL02", 6, 5),
+            ("LT01", 6, 6),
+            ("CP01", 8, 1),
+        ],
+        "test/fixtures/linter/whitespace_errors.sql": [
+            ("LT09", 1, 1),
+            ("LT01", 2, 9),
+            ("LT01", 3, 12),
+            ("LT02", 4, 1),
+            ("CP01", 6, 10),
+        ],
+    }
+
+
+@pytest.mark.parametrize(
+    "path,stats",
+    [
+        (
+            "multifile_a",
+            {
+                "avg per file": 2.5,
+                "clean": 0,
+                "clean files": 0,
+                "exit code": 111,
+                "files": 2,
+                "status": "FAIL",
+                "unclean": 2,
+                "unclean files": 2,
+                "unclean rate": 1.0,
+                "violations": 5,
+            },
+        ),
+        (
+            "multifile_b",
+            {
+                "avg per file": 2.0,
+                "clean": 0,
+                "clean files": 0,
+                "exit code": 111,
+                "files": 2,
+                "status": "FAIL",
+                "unclean": 2,
+                "unclean files": 2,
+                "unclean rate": 1.0,
+                "violations": 4,
+            },
+        ),
+    ],
+)
+def test__linter__linting_result_stats(path, stats):
+    """Test that a LintingResult can get the right stats with multiple files.
+
+    https://github.com/sqlfluff/sqlfluff/issues/5673
+    """
+    lntr = Linter()
+    result = lntr.lint_paths((f"test/fixtures/linter/exit_codes/{path}",))
+    # NOTE: We're using fake return codes for testing purposes.
+    assert result.stats(111, 222) == stats
 
 
 @pytest.mark.parametrize("processes", [1, 2])
@@ -382,18 +365,77 @@ def test__linter__linting_unexpected_error_handled_gracefully(
 
 
 def test__linter__empty_file():
-    """Test linter behaves nicely with an empty string."""
+    """Test linter behaves nicely with an empty string.
+
+    Much of this test is about making sure that ParsedString is
+    instantiated appropriately.
+    """
     lntr = Linter(dialect="ansi")
     # Make sure no exceptions raised and no violations found in empty file.
     parsed = lntr.parse_string("")
+    # There should still be a parsed variant
+    assert parsed.parsed_variants
+    assert len(parsed.parsed_variants) == 1
+    root_variant = parsed.parsed_variants[0]
+    # That root variant should still have a templated file and a parsed tree
+    # (although that parsed tree will likely just be an end of file marker).
+    assert root_variant.templated_file
+    assert root_variant.tree
+    # No violations
     assert not parsed.violations
 
 
+def test__linter__parse_fail():
+    """Test linter behaves as expected with an unparsable string.
+
+    Much of this test is about making sure that ParsedString is
+    instantiated appropriately.
+    """
+    lntr = Linter(dialect="ansi")
+    # Try and parse something which obviously isn't SQL
+    parsed = lntr.parse_string("THIS IS NOT SQL")
+    # There should still be a parsed variant
+    assert parsed.parsed_variants
+    assert len(parsed.parsed_variants) == 1
+    root_variant = parsed.parsed_variants[0]
+    # That root variant should still have a templated file and a parsed tree...
+    assert root_variant.templated_file
+    assert root_variant.tree
+    # ...but that tree should contain an unparsable segment.
+    assert "unparsable" in root_variant.tree.type_set()
+    # There *should* be violations because there should be a parsing fail.
+    assert parsed.violations
+    assert any(isinstance(v, SQLParseError) for v in parsed.violations)
+
+
+def test__linter__templating_fail():
+    """Test linter behaves as expected with invalid jinja template.
+
+    Much of this test is about making sure that ParsedString is
+    instantiated appropriately.
+    """
+    lntr = Linter(dialect="ansi")
+    # Try and parse something which breaks Jinja templating.
+    parsed = lntr.parse_string("{% if foo %}")
+    # For a templating fail, there won't be a parsed variant.
+    assert not parsed.parsed_variants
+    # There *should* be violations because there should be a templating fail.
+    assert parsed.violations
+    assert any(isinstance(v, SQLTemplaterError) for v in parsed.violations)
+
+
 @pytest.mark.parametrize(
-    "ignore_templated_areas,check_tuples",
+    "path,rules,ignore_templated_areas,check_tuples",
     [
-        (True, [("LT01", 3, 39), ("LT01", 3, 40)]),
         (
+            "test/fixtures/templater/jinja_h_macros/jinja.sql",
+            "L006",
+            True,
+            [("LT01", 3, 39), ("LT01", 3, 40)],
+        ),
+        (
+            "test/fixtures/templater/jinja_h_macros/jinja.sql",
+            "L006",
             False,
             [
                 # there are still two of each because LT01 checks
@@ -405,9 +447,23 @@ def test__linter__empty_file():
                 ("LT01", 3, 40),
             ],
         ),
+        (
+            "test/fixtures/linter/jinja_variants/simple_CP01.sql",
+            "CP01",
+            False,
+            [
+                # We should get violations from both sides of the if
+                # statement without doubling up on the one outside.
+                ("CP01", 2, 10),
+                ("CP01", 2, 34),
+                ("CP01", 2, 52),
+            ],
+        ),
     ],
 )
-def test__linter__mask_templated_violations(ignore_templated_areas, check_tuples):
+def test__linter__mask_templated_violations(
+    path, rules, ignore_templated_areas, check_tuples
+):
     """Test linter masks files properly around templated content.
 
     NOTE: this also tests deduplication of fixes which have the same
@@ -416,13 +472,13 @@ def test__linter__mask_templated_violations(ignore_templated_areas, check_tuples
     lntr = Linter(
         config=FluffConfig(
             overrides={
-                "rules": "L006",
+                "rules": rules,
                 "ignore_templated_areas": ignore_templated_areas,
                 "dialect": "ansi",
             }
         )
     )
-    linted = lntr.lint_path(path="test/fixtures/templater/jinja_h_macros/jinja.sql")
+    linted = lntr.lint_path(path=path)
     assert linted.check_tuples() == check_tuples
 
 
@@ -472,7 +528,7 @@ def test__linter__encoding(fname, config_encoding, lexerror):
             }
         )
     )
-    result = lntr.lint_paths([fname])
+    result = lntr.lint_paths((fname,))
     assert lexerror == (SQLLexError in [type(v) for v in result.get_violations()])
 
 
@@ -534,3 +590,69 @@ def test_normalise_newlines():
     in_str = "SELECT\r\n foo\n FROM \r \n\r bar;"
     out_str = "SELECT\n foo\n FROM \n \n\n bar;"
     assert out_str == Linter._normalise_newlines(in_str)
+
+
+@pytest.mark.parametrize(
+    "fix_even_unparsable",
+    [False, True],
+)
+def test_unparsable_fix_output(fix_even_unparsable):
+    """Tests functionality and logging output with unparsable sections.
+
+    NOTE: While we cover different paths, the result for this test is the
+    same for both values of `fix_even_unparsable`. We probably need a better
+    test case at some point so that we can actually see the difference.
+    """
+    config = FluffConfig(
+        overrides={"fix_even_unparsable": fix_even_unparsable, "dialect": "ansi"}
+    )
+    linter = Linter(config=config)
+    # Attempt to fix it, capturing the logging output.
+    with fluff_log_catcher(logging.WARNING, "sqlfluff.linter") as caplog:
+        result = linter.lint_paths(
+            ("test/fixtures/linter/parse_error_2.sql",),
+            fix=True,
+            apply_fixes=True,
+            fixed_file_suffix=f"_{fix_even_unparsable}_fix",
+            fix_even_unparsable=fix_even_unparsable,
+        )
+    # Assert that it parsed (i.e. we found a select_statement), but with an
+    # unparsable section in there too.
+    assert result.tree
+    assert "select_statement" in result.tree.descendant_type_set
+    assert "unparsable" in result.tree.descendant_type_set
+    # We should still find linting issues too
+    assert result.check_tuples(raise_on_non_linting_violations=False) == [
+        ("CP01", 2, 7),  # `a as b` - capitalisation of AS
+        ("AL03", 3, 5),  # 42 is an expression without an alias
+        # The unparsable section is (wrongly) detected as an indentation issue.
+        ("LT02", 4, 1),
+        ("CP01", 5, 1),  # `from` is uncapitalised
+    ]
+    # We should make sure that the warning that asks users to report a bug is
+    # NOT present. i.e. the warning which could happen in `lint_fix_parsed()`.`
+    assert "Please report this as a bug" not in caplog.text
+    # Also not the `fix not applied`. The one in `_warn_unfixable()`
+    assert "it would re-cause the same error" not in caplog.text
+    # In fact, there shouldn't be any warnings at all.
+    assert not caplog.text.strip()
+    # In both cases, the final capitalisation and the `a as b` sections should have
+    # been fixed (because they aren't in the unparsable section).
+    assert "from cte" not in result.tree.raw
+    assert "FROM cte" in result.tree.raw
+    assert "a as b" not in result.tree.raw
+    assert "a AS b" in result.tree.raw
+    # Check whether the file was persisted. If `fix_even_unparsable` was set, then
+    # there should be a file, and it should have the fixes from above in it. If not
+    # then there should be no fixed file, as the persist will have been aborted due
+    # to the parsing issues.
+    predicted_fix_path = (
+        f"test/fixtures/linter/parse_error_2_{fix_even_unparsable}_fix.sql"
+    )
+    if fix_even_unparsable:
+        with open(predicted_fix_path, "r") as f:
+            fixed_sql = f.read()
+        assert result.tree.raw == fixed_sql
+    else:
+        with pytest.raises(FileNotFoundError):
+            open(predicted_fix_path, "r")
